// The following is an example of implmenting a WR-28 Band Pass filter
// See the document "TLM GPU Usage.docx"  located in the DATA folder for a more detailed explanation. 

/*
    Copyright 2010 Computational Electromagnetics Research Laboratory (CERL)
    University of Victoria, Victoria, B.C. Canada
    http://www.cerl.ece.uvic.ca
    http://www.cerl.ece.uvic.ca/Open_Source_CEM.htm
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



// includes, system
#include <omp.h>
#include <iostream>
#include <math.h>


// Required to include CUDA vector types
#include <vector_types.h>
#include "cutil.h"

#include "TLMSim.h"
#include "TLM_GPU.h"





// Function Declaration

void CPU_SCN_3D_Sheet_HalfSine_Excitation_X_Prop(int x_min,int x_max,int y_min,int y_max,int z_min,int z_max, float excitation);
void CPU_SCN_Iterate(int numIterations);




////////////////////////////////////////////////////////////////////////////////
// Program main
// 
////////////////////////////////////////////////////////////////////////////////
int 
main(int argc, char** argv)
{   

	float scale = 1; //max:5.3 for Quadro FX 5600; 
	// instantiate GPU object
	GPU_Space::GPU gpuA("Data/STRUCTURE_DEFINITIONS.csv",scale);
	gpuA.iterations_MaxNum = 32000; // set up the maximum number of iterations 
	
	printf(" Mesh Dimensions: X = %d, Y = %d, Z = %d, Total Nodes: %d \n",gpuA.meshDim_X,gpuA.meshDim_Y,gpuA.meshDim_Z,gpuA.meshDim_X*gpuA.meshDim_Y*gpuA.meshDim_Z );
	
	printf(" Size of Mesh: %4.3f MBytes \n",float(12*gpuA.meshDim_X*gpuA.meshDim_Y*gpuA.meshDim_Z*sizeof(float))/1000000 );
	
	//////////////////////////////
	// REFERENCE EXCITATION setup
	// instantiate an excitation object strucutre (plane) with polerization and direction of propagation defined
	int excite_ID_0_REF = gpuA.excitation_CreateStructure(	1, // Propagation Direction: X_Dir=+/-1, Y_Dir=+/-2, Z_Dir=+/-3
															3, //Polerization: X=1, Y=2, Z=3
															4*scale, //X_min
															4*scale, // X_max
															2*scale-(scale-1), //Y_min
															34*scale, // Y_max
															2*scale-(scale-1), //Z_min
															18*scale);  // Z_max
	// Generate an excitation waveform (Gaussian Sine wave) and allocate to the above excitation object
	gpuA.excitation_generate_SineGaussian_array(excite_ID_0_REF , // Excitation Object ID
												31E9, //Freq (Hz)within Gaussian shape
												1,    // Magnitude of Gaussian shape
												200*scale, // Width of Gaussian Shape (in time steps)
												1000*scale,// Position of Gaussian Peak
												2500*scale);// Excitation Array size
	
	gpuA.excitation_Save(excite_ID_0_REF, // Excitation Object ID
						 "Data/Excitation_REF.csv"); // File name to save excitation array
	//
	//////////////////////////

	//////////////////////////////
	// FILTER EXCITATION setup
	int excite_ID_1_FILT = gpuA.excitation_CreateStructure(1,3,64*scale,64*scale,2*scale-(scale-1),34*scale,2*scale-(scale-1),18*scale);
	gpuA.excitation_generate_SineGaussian_array(excite_ID_1_FILT,31E9,1,200*scale,1000*scale,2500*scale);
	//
	//////////////////////////////
	
	//int excite_ID_2_TEST = gpuA.excitation_CreateStructure(1,3,72,72,2,34,2,18);  // inclusive co-ords
	//gpuA.excitation_generate_SineGaussian_array(excite_ID_2_TEST,31E9,1,150,600,gpuA.iterations_MaxNum);
	

	//////////////////////
	// SAMPLING SETUP  
	int sample_ID_0_REF = gpuA.sample_Probe_add(gpuA.iterations_MaxNum, // Max Number of samples to capture
												3, // Polerization Direction: X=1, Y=2, Z=2, NodeAverage=0
												12*scale,// X-coord
												17*scale,// Y-coord
												9*scale);// Z-coord
	int sample_ID_1_PROBE1 = gpuA.sample_Probe_add(gpuA.iterations_MaxNum,3,72*scale,17*scale,9*scale);
	int sample_ID_2_PROBE2 = gpuA.sample_Probe_add(gpuA.iterations_MaxNum,3,201*scale,17*scale,9*scale);//201 // 190
	//
	//////////////////////

	gpuA.iteration_timer_display_ENABLE = true; // enable the iteration timer to display status updates every two seconds
	gpuA.iteration_timer.start("GPU Iteration Timer Start"); // Start the iteration timer

	//gpuA.Mesh_Device_to_host_memcopy(); // retrieve GPU node stucuture results to CPU 
	//gpuA.Output_Node_Layer_XY_Plane("DATA/Output_Layer_Pre.csv",9,0);
	int capt_tmp = 30; // DEBUG: used to pause execution at a specific iteration count
	int impulse_idx = 26; // DEBUG: used to send a debug impulse wave at a specific iteration
	
	
	// Iteration for-next loop
	for ( int iter = 0 ; iter < gpuA.iterations_MaxNum; iter++){
		
		// Excitation execution: will excite a half sine guassian pulse defined for each excitation object
		gpuA.excitation_execute_HalfSine(excite_ID_0_REF); 
		gpuA.excitation_execute_HalfSine(excite_ID_1_FILT);

		///////////////
		// Debug impulse
		//if (iter == impulse_idx ) {
		//	gpuA.excitation_execute_HalfSine_Impulse(excite_ID_0_REF);
		//	/*for (int z_idx = 90 ; z_idx<= 110;z_idx++){	
		//		gpuA.excitation_execute_Impulse(132,20,z_idx,1.00);
		//		gpuA.excitation_execute_Impulse(100,100,z_idx,1.00);
		//	}
		//	*/
		//}

		//
		////////////////

		// SCATTER excitation (note: impulse iterchange is absorbed into scatter routine)
		gpuA.iteratate(); 
	
		  // Debug: Capture and Write to file one slice of the mesh
		//if (iter == capt_tmp){
		//	gpuA.Mesh_Device_to_host_memcopy(); // retrieve GPU node stucuture results to CPU 
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Avg.csv",9,0);
		//	/*gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_01.csv",9,1);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_02.csv",9,2);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_03.csv",9,3);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_04.csv",9,4);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_05.csv",9,5);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_06.csv",9,6);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_07.csv",9,7);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_08.csv",9,8);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_09.csv",9,9);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_10.csv",9,10);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_11.csv",9,11);
		//	gpuA.Output_Node_Layer_XY_Plane("Data/LayerData/LayerOutput_Arm_12.csv",9,12);*/
		//	int blank = 0;
		//}
		
		// Execute sampling: will call sample kernels and store on GPU until extracted at the end of run
		gpuA.sample_execute(sample_ID_0_REF);
		gpuA.sample_execute(sample_ID_1_PROBE1);
		gpuA.sample_execute(sample_ID_2_PROBE2);


		// BOUNDARY kernel calls. 
		for (int bndry_idx = 0 ; bndry_idx < gpuA.boundaries_Num ; bndry_idx++){
			gpuA.cuda_boundary_execute(bndry_idx);
		}
	
	}// end of iteration for-next
	
	// If the last interation index is odd then must itherchange all nodes in mesh back to propoer positions
	if(gpuA.iteration_index%2 == 1){ 
		gpuA.interchange_3D_mesh();
	}

	printf("\n");
	gpuA.iteration_timer.stop("TIMER STOP");
	printf("Elapsed Time: %f (sec) \n",gpuA.iteration_timer.elapsed_time());
	

	//gpuA.GPU_cuda_Test();
	
	///////////
	// Extract GPU Mesh from GPU to CPU
	gpuA.Mesh_Device_to_host_memcopy(); // retrieve GPU node stucuture results to CPU 

	// export a slice of mesh data (XY plane at z_slice) to a .csv file)
	gpuA.Output_Node_Layer_XY_Plane("DATA/Output_Layer_Post.csv",
									 9,// Z-Layer coord to extract
									 0); // Voltage to extract to .csv ('1-12', or '0' for average of all 12)
	//
	//////////
	
	//////////
	// Extract sample arrays from GPU and Save them to .csv files
	//  1st parameter = sample object ID
	//  2nd parameter = Number of samples to extract. If '0' then extracts ALL samples
	gpuA.sample_gpu_samples_get(sample_ID_0_REF,0); // Extract sample array from 
	gpuA.sample_gpu_samples_get(sample_ID_1_PROBE1,0);// Extract sample array from GPU
	gpuA.sample_gpu_samples_get(sample_ID_2_PROBE2,0);// Extract sample array from GPU
	
	gpuA.sample_save_to_file("DATA/Sampler_0_REF.csv",sample_ID_0_REF); // Save sample array to .csv file
	gpuA.sample_save_to_file("DATA/Sampler_1_Probe1.csv",sample_ID_1_PROBE1);// Save sample array to .csv file
	gpuA.sample_save_to_file("DATA/Sampler_2_Probe2.csv",sample_ID_2_PROBE2);// Save sample array to .csv file
	//
	//////////


	fprintf(stdout,"Job done, hit <enter> to exit\n");
	getc(stdin); // If you want to run Profiler then comment out this line 
	exit(0);

} // end of main()







