// TLM CUDA Relay Calling file
// This file contains CUDA code that is called from the Host. It manages calls to the GPU
// to: initialize, query, transfer memory, and launch GPU kernels


/*
    Copyright 2010 Computational Electromagnetics Research Laboratory (CERL)
    University of Victoria, Victoria, B.C. Canada
    http://www.cerl.ece.uvic.ca
    http://www.cerl.ece.uvic.ca/Open_Source_CEM.htm
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


// includes, project
#include <cufft.h>
#include <cutil.h>

// includes, kernels
#include "Kernel.cu"



  //////////////////////////
 // OOP related routines  /
//////////////////////////

extern "C" bool cuda_Device_Init_wID(int deviceID ){
	//cudaError_t 
	int numDevices ;
	cudaError_t cu_error = cudaGetDeviceCount (&numDevices);
	if (cu_error != cudaSuccess){
		printf("Error: Could not get Device Count\n");
		return false; 
	}

	if ( (deviceID < numDevices) && ( deviceID>=0 ) ){
		cudaSetDevice( deviceID );
		return true;
	} else if (deviceID == -1){	
		cu_error = cudaSetDevice(0);
		if (cu_error != cudaSuccess){
			printf("Error: Could not get Device Count\n");
			return false; 
		}else{
			int devtest;
			cudaGetDevice(&devtest);
			return true;
		}
	} else  {
		printf("Error: Device ID out of bounds\n");
		return false;
	}
        //cudaSetDevice( cutGetMaxGflopsDeviceId() );
}

extern "C" int cuda_NumOfDevices_get(){
	int numDevices ;
	cudaError_t cu_error = cudaGetDeviceCount (&numDevices);
	if (cu_error != cudaSuccess){
		printf("Error: Could not get Device Count\n");
		return 0; 
	}
	return numDevices;
}


extern "C" float* cuda_GPU_mesh_Mem_allocate(int meshDim_X,int meshDim_Y,int meshDim_Z){
	cudaDeviceProp devprop; 
	cudaError_t dev_err = cudaGetDeviceProperties(&devprop,0);

	/////////////////////////////////////////////
	// struct Grid_3D_nodes
    // allocate global device memory for struct Grid_3D_nodes

	unsigned int size_Grid_3D_nodes = meshDim_X*meshDim_Y*meshDim_Z; 
    unsigned int mem_size_Grid_3D_nodes = 12*sizeof(float) * size_Grid_3D_nodes;
	float* d_Grid_3D_nodes;
    CUDA_SAFE_CALL(cudaMalloc((void**) &d_Grid_3D_nodes, mem_size_Grid_3D_nodes));
	
    // Copy memory from Host to Device
    //CUDA_SAFE_CALL(cudaMemcpy(d_Grid_3D_nodes, Grid_3D_nodes , mem_size_Grid_3D_nodes,cudaMemcpyHostToDevice) );
	//
    ////////////////////////////
	return d_Grid_3D_nodes;
}

extern "C" void cuda_mesh_CPU_to_GPU_copy_relay(float* d_GPU_Device_Mesh, float* h_mesh_data, int meshDim_X,int meshDim_Y,int meshDim_Z){
	unsigned int size_Grid_3D_nodes = meshDim_X*meshDim_Y*meshDim_Z; 
    unsigned int mem_size_Grid_3D_nodes = 12*sizeof(float) * size_Grid_3D_nodes;
	CUDA_SAFE_CALL(cudaMemcpy(d_GPU_Device_Mesh, h_mesh_data, mem_size_Grid_3D_nodes,cudaMemcpyHostToDevice) );
}

extern "C" void cuda_mesh_CPU_to_GPU_get(Varm* d_GPU_Device_Mesh, Varm* h_mesh_data, int meshDim_X,int meshDim_Y,int meshDim_Z){
    unsigned int size_Grid_3D_nodes = meshDim_X*meshDim_Y*meshDim_Z; 
    unsigned int mem_size_Grid_3D_nodes = sizeof(Varm) * size_Grid_3D_nodes;
	CUDA_SAFE_CALL(cudaMemcpy( h_mesh_data, d_GPU_Device_Mesh ,mem_size_Grid_3D_nodes,cudaMemcpyDeviceToHost) );
}

extern "C" void cuda_Free_GPU_Mesh(Varm* GPU_Device_Mesh){
	cudaFree(GPU_Device_Mesh);
}

extern "C" void cuda_reflect_coef_constants_set_relay(float* reflection_coef_list, int boundaryNum){
	// TBD: boundaryNum is not the number of reflection_coef constantants, must be altered 
	/////////////////////////////////////////////
    // allocate CONSTANT device memory for r
    unsigned int size_ref = boundaryNum;
    unsigned int mem_size_ref = sizeof(float) * size_ref;
    // Copy memory from Host to Device
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_refl_coef,&reflection_coef_list[0],mem_size_ref));
	//
    ////////////////////////////
}
extern "C" void cuda_Mesh_Block_constants_set_relay(int meshDimX,int meshDimY ,int meshDimZ ,int thread_blk_X ,int thread_blk_Y ,int thread_blk_Z ){
	/////////////////////////////////////////////
    // allocate CONSTANT device memory for r
    unsigned int size = 1;
    unsigned int mem_size = sizeof(int) * size;
    // Copy memory from Host to Device
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_meshDimX,&meshDimX,mem_size));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_meshDimY,&meshDimY,mem_size));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_meshDimZ,&meshDimZ,mem_size));
	
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_thread_blk_X ,&thread_blk_X ,mem_size));// OBSOLETE
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_thread_blk_Y ,&thread_blk_Y ,mem_size));// OBSOLETE
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(C_thread_blk_Z ,&thread_blk_Z ,mem_size));// OBSOLETE

	//
    ////////////////////////////
}

// ITERATION relay call
extern "C" void cuda_iteration_relay(int iteration, float* GPU_Device_Mesh, iterationSetupType* iterSetup ){
		// Alternate between iteration kernel calls to accomodate both SCATTER AND IMPULSE INTERCHANGE routines
	if (iteration%2 == 0 )	{
		// SCATTER kernel call
		//mem_speed_test_kernel_11<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);}
		kernel_v2_TLM_3D_Iterate_scatter<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		cudaThreadSynchronize();
		//kernel_v2_TLM_3D_Iterate_scatter_64thread_1st_Pass<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		//kernel_v2_TLM_3D_Iterate_scatter_64thread_2nd_Pass<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);

	}else{
		// SCATTER and IMPULSE INTERCHANGE kernel call
		//mem_speed_test_kernel_11_part2<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);}
		//kernel_v2_TLM_3D_Iterate_scatter<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		kernel_v2_TLM_3D_Iterate_scatter_interchange<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		cudaThreadSynchronize();
		//kernel_v2_TLM_3D_Iterate_scatter_interchange_2_1st_Pass<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		//cudaThreadSynchronize();
		//kernel_v2_TLM_3D_Iterate_scatter_interchange_2_2nd_Pass<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		//cudaThreadSynchronize();
		//kernel_v2_TLM_3D_Iterate_scatter_interchange_MEM_TEST<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
	}
	

}


extern "C" void cuda_TLM_3D_InterchangeOnly(int iteration, float* GPU_Device_Mesh, iterationSetupType* iterSetup ){
		kernel_v2_TLM_3D_Iterate_InterchangeOnly<<<iterSetup->grid, iterSetup->threads>>>( GPU_Device_Mesh);
		cudaThreadSynchronize();
}

extern "C" void cuda_boundary_execute_relay(float* GPU_Device_Mesh, int Xmin,int Xmax,int Ymin,int Ymax,int Zmin,int Zmax, float reflect_coef){
	dim3 threads_11_bound(16,16,1);
	dim3 grid_11_bound;
	int tmpX=0; 
	int tmpY=0; 
	if (Xmin== Xmax){
		tmpX=(Ymax-Ymin)/16+1; 
		//if ((Ymax-Ymin)%16 != 0) tmpX++;
		tmpY=(Zmax-Zmin)/16+1; 
		//if ((Zmax-Zmin)%16 != 0) tmpY++;
		grid_11_bound.x= tmpX+1;
		grid_11_bound.y = tmpY+1;
		//mem_speed_test_kernel_11_Boundary<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin,Xmax,Ymin+1,Ymax,Zmin+1,Zmax, reflect_coef); 
		kernel_v2_TLM_3D_Boundary_2<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin,Xmax,Ymin+1,Ymax,Zmin+1,Zmax, reflect_coef); 
	}else if (Ymin== Ymax){
		tmpX=(Xmax-Xmin)/16+1; 
		//if ((Xmax-Xmin)%16 != 0) tmpX++;
		tmpY=(Zmax-Zmin)/16+1; 
		//if ((Zmax-Zmin)%16 != 0) tmpY++;
		grid_11_bound.x= tmpX+1;
		grid_11_bound.y = tmpY+1;
		//mem_speed_test_kernel_11_Boundary<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin+1,Xmax,Ymin,Ymax,Zmin+1,Zmax, reflect_coef); 
		kernel_v2_TLM_3D_Boundary_2<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin+1,Xmax,Ymin,Ymax,Zmin+1,Zmax, reflect_coef); 
	}else if (Zmin== Zmax){
		tmpX=(Xmax-Xmin)/16+1; 
		//if ((Xmax-Xmin)%16 != 0) tmpX++;
		tmpY=(Ymax-Ymin)/16+1; 
		//if ((Ymax-Ymin)%16 != 0) tmpY++;
		grid_11_bound.x= tmpX+1;
		grid_11_bound.y = tmpY+1;
		//mem_speed_test_kernel_11_Boundary<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin+1,Xmax,Ymin+1,Ymax,Zmin,Zmax, reflect_coef); 
		kernel_v2_TLM_3D_Boundary_2<<<grid_11_bound, threads_11_bound>>>(GPU_Device_Mesh,Xmin+1,Xmax,Ymin+1,Ymax,Zmin,Zmax, reflect_coef); 
	}


}

extern "C" void cuda_excitation_execute_relay(float* GPU_Device_Mesh, int scn_arm, int x_min,int x_max,int y_min,int y_max,int z_min,int z_max, float excitation, int iteration_index){
	int u_min, u_max; // span of half sine wave distribution
	int v_min, v_max; // polarization direction
	int w_plane; // propogation direction
	//ALL = 0, XnegYpol = 11, XposYpol = 3, XnegZpol = 10, XposZpol = 6, 
	//		   YnegXpol = 12, YposXpol = 1, YnegZpol =  7, YposZpol = 5, 
	//		   ZnegXpol =  9, ZposXpol = 2, ZnegYpol =  8, ZposYpol = 4
	
	// examples: XposYpol  = propogation in the postive  X direction with poloarization in the Y direction
	//           ZnegXpol  = propogation in the negative Z direction with poloarization in the X direction

	switch(scn_arm){
		case 0 :
				// default : no propagation direction, just span(Y) and polerization(Z) is set.  
				u_min = y_min;  // span of half sine wave distribution
				u_max = y_max;	// span of half sine wave distribution
				v_min = z_min;  // dir of polarization
				v_max = z_max;  // dir of polarization
				w_plane = x_min;  // direction of propogation
			break;
		
		case 1 :  //-ve Y prop, X polarization
		case 12 : //+ve Y prop, X polarization
				u_min = z_min; 
				u_max = z_max;	
				v_min = x_min;  
				v_max = x_max;  
				w_plane = y_min;  
			break;
		case 2 : //-ve Z prop, X polarization
		case 9 : //+ve Z prop, X polarization
				u_min = y_min; 
				u_max = y_max;	
				v_min = x_min;  
				v_max = x_max;  
				w_plane = z_min;  
			break;
		case 3 :  //-ve X prop, Y polarization
		case 11 : //+ve X prop, Y polarization
				u_min = z_min; 
				u_max = z_max;	
				v_min = y_min;  
				v_max = y_max;  
				w_plane = x_min;  
			break;
		case 4 : //-ve Z prop, Y polarization
		case 8 : //+ve Z prop, Y polarization
				u_min = x_min; 
				u_max = x_max;	
				v_min = y_min;  
				v_max = y_max;  
				w_plane = z_min;  
			break;
		case 5 : //-ve Y prop, Z polarization
		case 7 : //+ve Y prop, Z polarization
				u_min = x_min; 
				u_max = x_max;	
				v_min = z_min;  
				v_max = z_max;  
				w_plane = y_min;  
			break;
		case 6 :  //-ve X prop, Z polarization
		case 10 : //+ve X prop, Z polarization
				u_min = y_min;  // span of half sine wave distribution
				u_max = y_max;	// span of half sine wave distribution
				v_min = z_min;  // dir of polarization
				v_max = z_max;  // dir of polarization
				w_plane = x_min;  // direction of propogation
			break;
		default:
			break;
	}
		
	int blocksPerUGrid_Excite = (u_max - u_min )/20 + 1;
	//if ( (u_max - u_min )%20 != 0) blocksPerUGrid_Excite++;
	int blocksPerVGrid_Excite = (v_max - v_min )/20 + 1;
	//if ( (v_max - v_min )%20 != 0) blocksPerVGrid_Excite++;
	dim3 threadsE( 20, 20,1); 
	dim3 gridE(blocksPerUGrid_Excite, blocksPerVGrid_Excite);
	//kernel_TLM_3D_Plane_HalfSine_Excitation_new<<<gridE,threadsE>>>(GPU_Device_Mesh,scn_arm , u_min,u_max,v_min,v_max,w_plane, excitation,iteration_index%2);
	kernel_v2_TLM_3D_Plane_HalfSine_Excitation<<<gridE,threadsE>>>(GPU_Device_Mesh,scn_arm , u_min,u_max,v_min,v_max,w_plane, excitation,iteration_index%2);
	cudaThreadSynchronize();

}


extern "C" void cuda_excitation_execute_Impulse_relay(float* GPU_Device_Mesh, int x,int y,int z, float excitation, int iteration_index){
	//mem_speed_test_kernel_11_Impulse_Excitation<<<1,1>>>( GPU_Device_Mesh,x,y, z, excitation, iteration_index%2);
	kernel_v2_TLM_3D_Plane_Impulse_Excitation<<<1,1>>>( GPU_Device_Mesh,x,y, z, excitation, iteration_index%2);
}



extern "C" void cuda_sampling_execute_relay(float* GPU_Device_Mesh, float* GPU_Samples_ptr,  int sample_index, int polerize_dir, int x, int y,int z, int iteration_index){
	float sample_result;
	//kernel_TLM_Sample<<< iterSetup->grid, iterSetup->threads >>>(GPU_Device_Mesh, z_slice_idx , GPU_Samples_ptr, sample_index,  polerize_dir,x,y,z,iterSetup->nodes_per_Zslice);
	//mem_speed_test_kernel_11_Sample<<<1,1>>>( GPU_Device_Mesh,GPU_Samples_ptr,sample_index,polerize_dir,  x , y , z, (iteration_index+1)%2);
	kernel_v2_TLM_3D_Sample<<<1,1>>>( GPU_Device_Mesh,GPU_Samples_ptr,sample_index,polerize_dir,  x , y , z, (iteration_index+1)%2);
	cudaThreadSynchronize();
}


extern "C" void cuda_gpu_sampleArray_Allocate_relay(float** d_samples, int sample_numMax){
	
	/////////////////////////////////////////////
	// struct Grid_3D_nodes
    // allocate global device memory for struct Grid_3D_nodes
	float* tmp_ptr = NULL;
    unsigned int size_in_bytes = sizeof(float) * sample_numMax;
    CUDA_SAFE_CALL(cudaMalloc((void**) &tmp_ptr, size_in_bytes));
	*d_samples=tmp_ptr;
    // Copy memory from Host to Device
    //CUDA_SAFE_CALL(cudaMemcpy(d_Grid_3D_nodes, Grid_3D_nodes , mem_size_Grid_3D_nodes,cudaMemcpyHostToDevice) );
	//
    ////////////////////////////

	
	CUDA_SAFE_CALL( cudaMemset((void*)tmp_ptr, 0 , size_in_bytes));

	//size_in_bytes = sizeof(float) * sample_numMax;
	//CUDA_SAFE_CALL(cudaMemcpy( h_samples, tmp_ptr ,size_in_bytes ,cudaMemcpyDeviceToHost) );

}

extern "C" void cuda_Samples_get_relay(float* d_samples, float*  h_samples, int num_samples_to_get){
	unsigned int size_in_bytes ;
	size_in_bytes = sizeof(float) * num_samples_to_get;
	//float* h_tmp = (float*) malloc(sizeof(float)*num_samples_to_get );
	//float* d_tmp = NULL;
	//CUDA_SAFE_CALL(cudaMalloc((void**)&d_tmp, size_in_bytes));
	//CUDA_SAFE_CALL(cudaMemcpy( h_tmp, d_tmp,size_in_bytes ,cudaMemcpyDeviceToHost) );
	CUDA_SAFE_CALL(cudaMemcpy( h_samples, d_samples ,size_in_bytes ,cudaMemcpyDeviceToHost) );
}




extern "C" void cuda_Device_to_host_memcopy_relay(float* d_data, float*  h_data, int num_nodes_to_get){
	cudaDeviceProp devprop; 
	cudaError_t dev_err = cudaGetDeviceProperties(&devprop,0);
	unsigned int size_in_bytes = 12*sizeof(float) * num_nodes_to_get;
	CUDA_SAFE_CALL(cudaMemcpy( h_data, d_data ,size_in_bytes ,cudaMemcpyDeviceToHost) );
	//cudaError_t cu_error = cudaMemcpy( h_data, d_data ,size_in_bytes ,cudaMemcpyDeviceToHost) ;
}

extern "C" void
cuda_save_results_to_CSVfile(char* filename, float samples_CUDA[], int sample_len, float delta_time){
	//FILE *OutFile = fopen("NormFFTData.csv","w");
	FILE *OutFile = fopen(filename,"w");

	if (!OutFile)
	{
		fprintf(stderr, "Error opening %s\nHit <Enter> to terminate the program\n", filename);
		fgetc(stdin);
		exit(0);
	}

	//fprintf(OutFile,"Grid=%dx%d , iterations=,%d  \n", V_X_SIZE,V_Y_SIZE,NUMOFITERATIONS  );
	fprintf(OutFile,"Index, Time(sec) ,samples_CUDA , Num of Samples =,%d \n",sample_len);
	
	float delta_Freq = 1/delta_time; //FREQ_SAMPLING/sample_len;
	
	for (int i = 0; i < sample_len; i++) {
		//Send data to file
		//fprintf(OutFile,"%f , %f , %f \n",samples_CUDA[i],fft_results[i]); // write: Time,fft_mag, fft_phase
		float sample_time = ((float)i)*delta_time;
		fprintf(OutFile,"%d, %E , %E  \n",i, sample_time, samples_CUDA[i]); 
	}
	//Close the file 
	fclose(OutFile);

	//
	///////////

}

extern "C" void cuda_Test(){
   	unsigned int size_in_bytes ;
	cudaError_t cu_err;
	cudaDeviceProp devprop; 
	cu_err = cudaGetDeviceProperties(&devprop,0);
	int numDevices;
	cu_err = cudaGetDeviceCount (&numDevices);

	//cu_err = cudaSetDevice(0);

	int num_samples_to_get = 1000;
	float* data_orig = (float*) malloc(sizeof(float)*num_samples_to_get );
	float* data_results = (float*) malloc(sizeof(float)*num_samples_to_get );
	for(int idx = 0 ; idx< num_samples_to_get ; idx++){
		data_orig[idx] = idx + 1;
		data_results[idx] = 0;
	}

	float* tmp_ptr = NULL;
    size_in_bytes = sizeof(float) * num_samples_to_get;
    CUDA_SAFE_CALL(cudaMalloc((void**)&tmp_ptr, size_in_bytes));
	//cu_err = cudaMalloc((void**)&tmp_ptr, size_in_bytes);
	
	CUDA_SAFE_CALL(cudaMemcpy( tmp_ptr,data_orig ,size_in_bytes ,cudaMemcpyHostToDevice) );
	
	
	CUDA_SAFE_CALL(cudaMemcpy( data_results, tmp_ptr ,size_in_bytes ,cudaMemcpyDeviceToHost) );

}


	
